// Copyright 2003, Philip Kohn

#include "qdocview.h"
#include <qpe/timeconversion.h>
#include <qpe/stringutil.h>
#include <qfile.h>
#include <qfileinfo.h>

#define assert(X)

QDocNote::QDocNote()
{
  clear();
}

QDocNote::~QDocNote()
{
  clear();
}

void
QDocNote::clear()
{
  m_firstWord = m_lastWord = -1;
  m_page = -1;
  m_start = QPoint(0,0);
  m_end = QPoint(0,0);
  m_date = QDateTime::currentDateTime();
  m_note = QString("");
  m_regionWords = QRegion();
  m_filename = "n/a";
}


void
QDocNote::save(QString &str)
{
  //  assert(m_filename != "n/a" && !m_filename.isEmpty());
  str += "<note file=\"" + Qtopia::escapeString(m_filename) + "\"";
  str += " page=\"" + QString::number(m_page) + "\"";
  str += " x0=\"" + QString::number(m_start.x()) + "\"";
  str += " y0=\"" + QString::number(m_start.y()) + "\"";
  str += " x1=\"" + QString::number(m_end.x()) + "\"";
  str += " y1=\"" + QString::number(m_end.y()) + "\"";
  str += " date=\"" + QString::number(TimeConversion::toUTC(m_date)) + "\"";
  str += ">\n";
  str += Qtopia::escapeString(m_note); // + "\n";
}

int
QDocNote::load(const QString &str, int index)
{
  XmlParser parser;
  index = parser.parse(str, index);
  clear();
  if (!parser.checkHead("note")) {
    printf("note header bad\n");
    return(-1);
  }
  if (!parser.get("page").isEmpty()) {
    m_page = parser.get("page").toInt();
  }
  if (!parser.get("x0").isEmpty()) {
    m_start.setX(parser.get("x0").toInt());
  }
  if (!parser.get("y0").isEmpty()) {
    m_start.setY(parser.get("y0").toInt());
  }
  if (!parser.get("x1").isEmpty()) {
    m_end.setX(parser.get("x1").toInt());
  }
  if (!parser.get("y1").isEmpty()) {
    m_end.setY(parser.get("y1").toInt());
  }
  if (!parser.get("date").isEmpty()) {
    m_date = TimeConversion::fromUTC(parser.get("date").toInt());
  }
  if (!parser.get("note").isEmpty()) {
    m_note = parser.get("note");
  } else {
    int cp =  str.find('<', index);
    //    printf("find < cp = %d\n", cp);
    if (cp < 0) {
      cp = str.length();
    }
    //    printf("cp = %d index %d\n", cp, index);
    //    assert(str[index-1] == '>');
    //    assert(str[index] == '\n');

    index++;
    if (cp > index) {
      m_note = XmlParser::unescape(str.mid(index, cp - index));
      //      printf("cp %d index %d str = %s\n", cp, index, m_note.latin1());
    } else {
      m_note = "";
    }
    index = cp;
  }
  if (parser.get("file").isEmpty()) {
    m_filename = "";
  } else {
    m_filename = XmlParser::unescape(parser.get("file"));
  }
  //  printf("loaded note %d: page %d text %s\n", index, m_page, m_note.latin1());
  return(index);
}


QDocNotes::QDocNotes()
{
  m_clean = true;
  m_noteSeperator.fill('=',43);
  m_noteSeperator += '\n';
  m_notes.clear();
  m_notes.setAutoDelete(true);
  m_locked = false;
}

QDocNotes::~QDocNotes()
{
  clear();
}

void
QDocNotes::clear()
{
  assert(m_clean);
  assert(!m_locked);
  m_notes.clear();
  m_text = NULL;
  m_setup_state.clear();
  m_paths = QStringList();
  m_clean = true;
  m_filepath = QString();
  m_width = m_height = 0;
  m_load_directory_path = QString();
}

void
QDocNotes::setupPage(QDocText *text)
{
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;

  m_width = text->m_state.width; 
  m_height = text->m_state.height;

  m_text = text;
  m_setup_state = text->m_state;

  QString filename = text->m_state.filename();
  while( (note = it.current()) != NULL ) {
    ++it;
    if (note->m_page == text->m_state.page
	&& (note->m_filename.isEmpty()
	    || note->m_filename == filename)) {
      note->m_firstWord = text->findWordPDF(absolutePosition(note->m_start));
      note->m_lastWord = text->findWordPDF(absolutePosition(note->m_end));
      updateQuote(note);
      if (note->m_firstWord >= 0 && note->m_lastWord >= 0) {
	note->m_regionWords = text->wordRegion(note->m_firstWord, note->m_lastWord);
      } else {
	note->m_regionWords = QRegion();
      }
    } else {
      note->m_firstWord = note->m_lastWord = -1;
      note->m_regionWords = QRegion();
    }
  }
}

// max width of 2048 = 11 bits
QPoint
QDocNotes::absolutePosition(QPoint pos)
{
  QPoint apos;
  apos.setX((pos.x() * m_width) >> 18);
  apos.setY((pos.y() * m_height) >> 18);
  //  printf("abs(%d,%d) = %d,%d (screen %d x %d)\n",
  // apos.x(), apos.y(), pos.x(), pos.y(), m_width, m_height);
  return(apos);
}

QPoint
QDocNotes::relativePosition(QPoint pos)
{
  QPoint rpos;
  rpos.setX((pos.x() << 18) / m_width);
  rpos.setY((pos.y() << 18) / m_height);
  //  printf("rel(%d,%d) = %d,%d (screen %d x %d)\n",
  // rpos.x(), rpos.y(), pos.x(), pos.y(), m_width, m_height);
  return(rpos);
}


QDocNote *
QDocNotes::find(QDocState &pos)
{
  return(find(pos.page, QPoint(pos.x, pos.y)));
}

QDocNote *
QDocNotes::find(int page, QPoint point)
{
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;
  while( (note = it.current()) != NULL ) {
    ++it;
    if (note->m_page == page 
	&& note->m_regionWords.contains(point)) {
      return(note);
    }
  }
  return(NULL);
}

QDocNote *
QDocNotes::find(int page, int word, bool closest)
{
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;
  int closestWord = 1000000;
  QDocNote *closestNote = NULL;
  int closestN = -1;
  int n = 0;

  while( (note = it.current()) != NULL ) {
    ++it; ++n;
    if (note->m_page == page) {
	if (word >= note->m_firstWord && word <= note->m_lastWord) {
	  //	  printf("found note %d from %d to %d\n", n,
	  // note->m_firstWord, note->m_lastWord);
	  return(note);
	}
	// this is very approximate, esp. for big ranges it is really bad
	// but is only used to put up some note on that page
	// when switching into note view
	if (QABS(word - note->m_firstWord) < closestWord) {
	  closestWord = QABS(word - note->m_firstWord);
	  closestNote = note;
	  closestN = n;
	}
	if (QABS(word - note->m_lastWord) < closestWord) {
	  closestWord = QABS(word - note->m_lastWord);
	  closestNote = note;
	  closestN = n;
	}
    }
  }
  if (closest && closestNote) {
    //    printf("found closest note %d from %d to %d\n",
    //	   closestN, closestNote->m_firstWord, closestNote->m_lastWord);
    return(closestNote);
  }
  return(NULL);
}

QDocNote *
QDocNotes::find(int page, int firstWord, int lastWord)
{
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;
  int n = 0;

  while( (note = it.current()) != NULL ) {
    ++it; ++n;
    if (note->m_page == page) {
	if (firstWord == note->m_firstWord && lastWord == note->m_lastWord) {
	  return(note);
	}
    }
  }
  return(NULL);
}


QRegion
QDocNotes::highlights(int page)
{
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;
  QRegion region; 
  while( (note = it.current()) != NULL ) {
    ++it;
    if (note->m_page == page && ! note->m_regionWords.isNull()) {
      region |= note->m_regionWords;
    }
  }
  return(region);
}

QDocNote*
QDocNotes::add(int page, QPoint start, QPoint end, const QString &noteString, QDocText *text)
{
  // don't add same note twice
  if (m_locked) {
    printf("@@@@@ not adding note because notes locked\n");
    return(NULL);
  }
  QDocNote *note = new QDocNote();
  note->m_page = page;
  note->m_firstWord = text->findWord(start);
  note->m_lastWord = text->findWord(end);
  note->m_start = relativePosition(text->wordBoxPDF(note->m_firstWord).center());
  note->m_end = relativePosition(text->wordBoxPDF(note->m_lastWord).center());
  note->m_date = QDateTime::currentDateTime();
  note->m_note = noteString;
  note->m_filename = text->m_state.filename();
  assert(!text->m_state.filename().isNull());
  QFileInfo fi(m_filepath);
  printf("add note text filename=%s notes filepath=%s  filename=%s\n",
	 text->m_state.basename().latin1(), m_filepath.latin1(), fi.baseName().latin1());
  // maybe should compare paths as well
  assert(text->m_state.basename() == fi.baseName());
  //  printf("add note [%s] [%s]\n", noteString.left(3).latin1(), noteString.right(3).latin1());
  note->m_regionWords = text->wordRegion(note->m_firstWord, note->m_lastWord);
  if (add(note)) {
    m_clean = false;
    printf("add note total=%d (count %d)\n", m_info.m_num_notes,
	   m_notes.count());
    m_info.m_num_notes++;
    m_info.notesChanged(this);
    return(note);
  } else {
    return(NULL);
  }
  //  printf("<<<<<<<<<<<add note not clean>>>>>>>\n");
}

void
QDocNotes::checkdups()
{
  int i, j;
#if 0
  for(i=0; i<m_notes.count(); i++) {
    for(j=i+1; j<m_notes.count(); j++) {
      QDocNote *no, *no2;
      no = m_notes.at(i);
      no2 = m_notes.at(j);
      if ((no->m_page == no2->m_page)
	  && (no->m_filename == no2->m_filename)
	  //	&& (no->m_start == no2->m_start)
	  //	&& (no->m_end == no2->m_end)
	  && (no->m_note == no2->m_note)) {

	printf("----\ncheckdups dup[%d,%d] %d==%d %s==%s   %d==%d %d==%d  %d==%d %d==%d   %s==%s\n",
	       i, j,
	       no->m_page,no2->m_page,
	       no->m_filename.latin1(), no2->m_filename.latin1(),
	       no->m_start.x(), no2->m_start.x(),
	       no->m_start.y(), no2->m_start.y(),
	       no->m_end.x(), no2->m_end.x(),
	       no->m_end.y(), no2->m_end.y(),
	       no->m_note.latin1(), no2->m_note.latin1());
	
	printf("********* dup note *********\n");
	assert(0);
      }
    }
  }
#endif
}

bool
QDocNotes::add(QDocNote *note)
{
  int i;

  for(i=0; i<m_notes.count(); i++) {
    QDocNote *no;
    no = m_notes.at(i);
#if 0
    printf("----\nadd dup[%d] %d==%d %s==%s   %d==%d %d==%d  %d==%d %d==%d   %s==%s\n",
	   i,
	   no->m_page,note->m_page,
	   no->m_filename.latin1(), note->m_filename.latin1(),
	   no->m_start.x(), note->m_start.x(),
	   no->m_start.y(), note->m_start.y(),
	   no->m_end.x(), note->m_end.x(),
	   no->m_end.y(), note->m_end.y(),
	   no->m_note.latin1(), note->m_note.latin1());
#endif

    if ((no->m_page == note->m_page)
	&& (no->m_filename == note->m_filename)
	//	&& (no->m_start == note->m_start)
	//	&& (no->m_end == note->m_end)
	&& (no->m_note == note->m_note)) {
      //      printf("********* dup note *********\n");
      delete note;
      return(false);
    }
  }

  //  printf("********* ADD non-dup note *********\n");
  m_notes.append(note);
  return(true);
}


QString
QDocNotes::toText(int page, QDocNote *selected_note, int *selected_note_position)
{
  QString str;
  QListIterator<QDocNote> it( m_notes );
  QDocNote *note;
  int ref_num = 1;

  while( (note = it.current()) != NULL ) {
    if (page < 0 || note->m_page == page) {
      str += m_noteSeperator;
      str += QString("[%1] page %2 ").arg(ref_num).arg(note->m_page);
      // a problem here is that if file is started with "open notes"
      // we don't really have a state yet.
      if (!m_text || !note->isSameFile(m_text->m_state.filename())) {
	str += QString("of '%1' ").arg(note->m_filename);
      }
      str += QString("(%1)\n").arg(note->m_date.toString());
      if (note == selected_note) {
	if (selected_note_position) {
	  *selected_note_position = str.length();
	}
      }
      str += note->m_note;
      // insure at least one return at end of note
      if (str.right(1) != "\n") {
	printf("**********toText adding return [%s]\n", str.right(3).latin1());
	str += "\n";
      }
    }
    ++it;
    ++ref_num;
  }
  return(str);
}

bool
QDocNotes::fromText(const QString &str, int noteCharPos, 
		    QDocNote **note)
{
  int index = 0;
  int len = str.length();
  int noteStart = -1;
  int textStart;
  int textEnd;
  int ref_num;
  bool ok;
  int i;
  QList<QDocNote> new_notes;
  QDocNote *new_note;
  bool bad = false;
  int noteCount = 0;


  assert(!m_locked);

  //  printf("fromText note %x\n", note);
  if (note) {
    //    printf("set note NULL\n");
    *note = NULL;
  }
  while(index < len) {
    noteStart = index;
    index = str.find("=======", index);
    if (index < 0) {
      break;
    }
    // start of ref line
    i = str.find('\n', index) + 1;
    if (i < 0 || str[i] != '[') {
      //      printf("did not find [, found '%c' in %s\n", (char)str[index], str.latin1());
      if (bad) break;
      bad = true;
      continue;
    }
    index = i;
    i = str.find(']', index+1);
    if (i < 0) {
      //      printf("did not find ]\n");
      if (bad) break;
      bad = true;
      continue;
    }
    ref_num = str.mid(index+1, i-index-1).toInt(&ok);
    if (! ok || ref_num < 1 || ref_num > m_notes.count()) {
      //      printf("bad ref count %d (%s)\n", ref_num,
      //	     str.mid(index+1, i-index-1).latin1());
      if (bad) break;
      bad = true;
      continue;
    }
    ref_num -= 1;
    textStart = str.find('\n', index) + 1;
    textEnd = str.find("=======", index) - 1;
    if (textStart < 0) {
      //      printf("no return text start found\n");
      if (bad) break;
      bad = true;
      textStart = index;
    }
    if (textEnd < 0) {
      textEnd = str.length() - 1;
    }
    index = textEnd;
#if 0
    printf("====\nparsed note %d:\ntext = %s\n====\n",
	   ref_num, 
	   str.mid(textStart, textEnd - textStart + 1).latin1());
#endif

    if (m_notes.at(ref_num)->m_page > 0) {

      new_note = new QDocNote;
      *new_note = *m_notes.at(ref_num);
      m_notes.at(ref_num)->m_page = -1;  // don't allow same note to be duped, note count gets off!
      new_note->m_note = str.mid(textStart, textEnd - textStart + 1);
      if (new_note->m_note.right(1) != "\n") {
	printf("fromText adding return [%s]\n", new_note->m_note.right(3).latin1());
	new_note->m_note += "\n";
      }
      updateQuote(new_note);
      new_notes.append(new_note);
      if (new_note->isSameFile(m_filepath)) {
	noteCount++;
      }

      if (noteCharPos >= noteStart 
	  && noteCharPos <= textEnd + 1) {
#if 0
	printf("found match for page %d, words %d to %d\n", 
	       new_note->m_page,
	       new_note->m_firstWord,
	       new_note->m_lastWord);
#endif
	if (note) {
	  //	printf("set note %x\n", new_note);
	  *note = new_note;
	}
      }
    }

  }

  if (new_notes.count() != m_notes.count()) {
    m_clean = false;
  } else {
    for(ref_num=0; ref_num<m_notes.count(); ref_num++) {
      // maybe should compare other things??
      if (new_notes.at(ref_num)->m_note != m_notes.at(ref_num)->m_note) {
	m_clean = false;
	break;
      }
    }
  }
  m_notes.clear();
  m_notes = new_notes;
  if (! m_clean) {
    //    printf("<<<<<<from text - not clean>>>>>\n");
    checkdups();
    printf("change num_notes to %d (total %d file %s)\n", 
	   noteCount, m_notes.count(), m_filepath.latin1());
    m_info.m_num_notes = noteCount;
    m_info.notesChanged(this);
    return(true);
  }
  return(false);
}

void
QDocNotes::updateQuote(QDocNote *note)
{
    // here is an attempt to change the range of the document quote based
    // on the text between double quotes in the note
    // NOTE: really need to be sure there are no double quotes in quotation from document!

  checkdups();

    if (m_text 
	&& note->isSameFile(m_text->m_state.path)
	&& note->m_page == m_text->m_state.page
	&& note->m_firstWord >= 0 
	&& note->m_lastWord >= 0) {
      int quoteStart, quoteEnd;
      quoteStart = note->m_note.find("[\"")+2;
      quoteEnd = note->m_note.find("\"]", quoteStart+2)-1;
      if (quoteStart >= 0 && quoteStart < 5 && quoteEnd > 0) {
	QString quote = note->m_note.mid(quoteStart, quoteEnd - quoteStart + 1);
	QString doc_quote = m_text->getText(note->m_firstWord, note->m_lastWord);
	if (quote != doc_quote) {
	   printf("extracted quote = %s\n", quote.latin1());
	   printf("extracted quote from document(%d,%d) = %s\n", 
		  note->m_firstWord, note->m_lastWord, doc_quote.latin1());
	  // look for case of quote is a subset of doc_quote
	  int find_start = doc_quote.find(quote);
	  if (find_start >= 0) {
	    //	    printf("found match at char %d\n", find_start);
	    int new_start = note->m_firstWord;
	    int new_end = note->m_lastWord;
	   if (new_end >= new_start
		  && m_text->getText(new_start,new_end).find(quote) >= 0) {
	     while(new_end >= new_start
		   && m_text->getText(new_start,new_end).find(quote) >= 0) {
	       new_end--;
	       printf("take away last word => %s\n", m_text->getText(new_start,new_end).latin1());
	     }
	     new_end++;
	   }
	   if (new_end >= new_start
	       && m_text->getText(new_start,new_end).find(quote) >= 0) {
	     //	     printf("try to remove from start\n");
	     while(new_end >= new_start
		   && m_text->getText(new_start,new_end).find(quote) >= 0) {
	       new_start++;
	       printf("take away first word => %s\n", m_text->getText(new_start,new_end).latin1());
	     }
	     new_start--;
	   }
	   QString new_quote = m_text->getText(new_start, new_end);
	   while (new_quote.left(1) == "\n") {
	     printf("addNote trim starting return\n");
	     new_quote.remove(0,1);
	   }
	   while (new_quote.right(1) == "\n") {
	     printf("addNote trim ending return\n");
	     new_quote.remove(new_quote.length()-1,1);
	   }
	   printf("==> new quote(%d,%d) is %s\n", new_start, new_end, new_quote.latin1());
	   QString new_note;
	   new_note = note->m_note;
	   new_note.replace(quoteStart, quoteEnd - quoteStart + 1,
			    new_quote);
	   if (note->m_note != new_note) {
	     printf("*** note changed!!\n");
	     note->m_note = new_note;
	     m_clean = false;
	   }
	   note->m_firstWord = new_start;
	   note->m_lastWord = new_end;
	   note->m_start = relativePosition(m_text->wordBoxPDF(note->m_firstWord).center());
	   note->m_end = relativePosition(m_text->wordBoxPDF(note->m_lastWord).center());
	  }
	}
      }
    }

  checkdups();

}

QString &
QDocNotes::getPath()
{
  return(m_filepath);
}


void
QDocNotes::setPath(const QString &path)
{
  assert(!m_locked);
  if (!m_load_directory_path.isNull()) {
    QFileInfo fi(path);
    printf("notes setPath(%s) clear notes if: loaded dir %s != %s\n",
	   path.latin1(), m_load_directory_path.latin1(), fi.dirPath(true).latin1());
    if (m_load_directory_path != fi.dirPath(true)) {
      printf("*** change loaded dir ****\n");
      m_load_directory_path = QString();
      clear();
    }
  } else {
    clear();  // could take this out to accumlate note "history" of files
  }
  m_filepath = path;
}

QString &
QDocNotes::getLoadDirectory()
{
  return(m_load_directory_path);
}

void
QDocNotes::setLoadDirectory(const QString &path) 
{
  m_load_directory_path = path;
  QFileInfo fi(m_filepath);
  if (fi.dirPath(true) != path) {
    printf("loadDirectory for dir %s not current dir %s\n",
	   path.latin1(), fi.dirPath(true).latin1());
  }
  assert(fi.dirPath(true) == path);
}


bool
QDocNotes::save(QDocState *state)
{
  return(save(m_filepath, state));
}


#ifdef ARM
extern int errno;
extern char *strerror(int);
#else
extern "C" {
  int errno;
  extern char *strerror(int);
}
#endif


bool
QDocNotes::save(const QString &filepath, QDocState *state)
{

  assert(!m_locked);
  m_locked = true;

  checkdups();

  assert(!filepath.isEmpty());
  QString newFilepath = filepath + ".tmp";
  QFileInfo fileinfo(filepath);
  QString filename = fileinfo.fileName();
  QString filedir = fileinfo.dirPath(true);
  int noteCount, noteCount2;

  // keep this assert since it protects against
  // writing over any none xml file!
  assert(fileinfo.extension(false) == "xml");

  // if you make the number of chars before  <info> more than 100
  // be sure to change QDocInfo::quickRead
  QString buf( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" );
  int ref_num = 1;
  int total_written;
  QCString str;
  QListIterator<QDocNote> it( m_notes );

  QFile f( newFilepath );

  QFileInfo path(state->path);

  QFileInfo notes_filepath(m_filepath);
  assert(notes_filepath.baseName() == path.baseName());

  if (fileinfo.baseName() != path.baseName()) {
    printf("QDocNotes::save ERROR filepath %s state->path %s\n",
	   filename.latin1(), path.baseName().latin1());
  }
  // insure we save note to same file as pdf
  assert(fileinfo.baseName() == path.baseName());

  printf("save %s page %d path %s\n",
  	 filepath.latin1(), state->page, state->path.latin1());

  if ( !f.open( IO_WriteOnly ) ) {
    printf("can not open %s\n", newFilepath.latin1());
    goto write_error;
  }

  buf += "<!DOCTYPE QPDFNOTES><QPDFNOTES>\n";

  noteCount = 0;
  QDocNote *note;
  while( (note = it.current()) != NULL ) {
    if (note->isSameFile(filepath)) {
      noteCount++;
    }
    ++it;
  }
  it.toFirst();

  if (m_info.m_num_notes != noteCount) {
    printf("!!!!!!! info notes %d count %d !!!!!!!!\n", m_info.m_num_notes, noteCount);
    //    m_info.m_num_notes = noteCount;
  }
  assert(m_info.m_num_notes == noteCount);

  m_info.save(buf);

  if (state) {
    buf += "<bookmarks>\n";
    state->save(buf);
    buf += "</bookmarks>\n";
  }

  str = buf.utf8();
  total_written = f.writeBlock( str.data(), str.length() );
  if ( total_written != int(str.length()) ) {
    goto write_error;
  }

  buf = "<paths>\n" + filedir + "\n</paths>\n";
  buf += "<notes>\n";
  str = buf.utf8();
  total_written = f.writeBlock( str.data(), str.length() );
  if ( total_written != int(str.length()) ) {
    goto write_error;
  }

  noteCount2 = 0;
  while( (note = it.current()) != NULL ) {
    buf = "";
    //    printf("save note file %s into %s?\n",
    //	   note->m_filename.latin1(), m_filepath.latin1());
    if (note->isSameFile(filepath)) {
      //      printf("yes\n");
      note->save(buf);
      str = buf.utf8();
      total_written = f.writeBlock( str.data(), str.length() );
      noteCount2++;
      if ( total_written != int(str.length()) ) {
	goto write_error;
      }
    }
    ++it;
  }

  assert(noteCount == noteCount2);

  buf = "</notes>\n</QPDFNOTES>\n";
  str = buf.utf8();
  total_written = f.writeBlock( str.data(), str.length() );
  if ( total_written != int(str.length()) ) {
    goto write_error;
  }
  f.close();
  
  if ( ::rename( newFilepath, m_filepath ) < 0 ) {
    qWarning( "problem renaming file %s to %s errno %d",
	      newFilepath.latin1(), m_filepath.latin1(), errno  );
    QFile::remove( newFilepath );
    goto write_error;
  }
  //  printf("<<<<<<<<save notes worked>>>>>>>>\n");
  m_clean = true;
  m_locked = false;

  checkdups();

  return true;

 write_error:

    printf("@@@@ notes write error on %s new %s (%s)\n", m_filepath.latin1(), 
	   newFilepath.latin1(),
	   strerror(errno));
    f.close();
    QFile::remove( newFilepath );
    m_locked = false;
    QMessageBox::critical( qApp->mainWidget(), QObject::tr( "Error writing notes" ),
			   QObject::tr( "QPdf was unable to\n"
					"save your notes.\n"
					"\"") + QString(strerror(errno)) + "\"");
    //					"Check space and permissions.\n"));
    printf("<<<<<<<<<<<<<<<<save notes failed>>>>>>>>>>\n");
    return false;
}



bool
QDocNotes::loadState(const QString &filepath, QDocState &state) 
{
  QFile f( filepath );
  QFileInfo fi(filepath);

  if (filepath.isNull()) {
    return false;
  }
  if ( !f.open(IO_ReadOnly) )
    return false;

  QString xml = QString(f.readAll());
  f.close();

  int index;

  index = xml.find("<bookmark ");
  if (index >= 0) {
    state.load(xml, index);
    printf("loadState = ");
    state.dump();
    return true;
  } else {
    printf("no bookmark\n");
    return false;
  }
}

bool
QDocNotes::load(QDocState *state)
{
  return(load(m_filepath, state));
}


bool
QDocNotes::load(const QString &filepath, QDocState *state)
{

  QFile f( filepath );
  QFileInfo fi(filepath);

  assert(!m_locked);
  m_locked = true;

  checkdups();

  m_info.clear();
  if (state) {
    state->clear();
    state->path = filepath;
  }

  if (filepath.isNull()) {
    m_locked = false;
    return false;
  }
  if ( !f.open(IO_ReadOnly) ) {
    m_locked = false;
    return false;
  }

  QString xml = QString(f.readAll());
  f.close();

  int index;

  index = xml.find("<info ");
  if (index >= 0) {
    m_info.load(xml, index);
  } else {
    printf("no info\n");
  }

  if (state) {
    index = xml.find("<bookmark ");
    if (index >= 0) {
      state->load(xml, index);
    } else {
      printf("no bookmark\n");
    }
  }

  // remove notes from this file
  // this is here more to allow multiple autotesters to run
  // without getting confused if one of them changes a note
  // in a directory of loaded notes
  QDocNote *note;
  for(int i = m_notes.count()-1; i >= 0; i--) {
    if (m_notes.at(i)->isSameFile(filepath)) {
      m_notes.remove(i);
    }
  }

  index = 0;
  int noteCount = 0;
  printf("loading notes...\n");
  for(;;) {
    XmlParser parser;
    index = xml.find("<note ", index);
    if (index < 0) {
      break;
    }
    note = new QDocNote;
    index = note->load(xml, index);
    if (index >= 0) {
      printf("got note file %s while loading path %s\n",
	     note->m_filename.latin1(),
	     m_filepath.latin1());
      if (note->m_filename.isEmpty()) {
	printf("note filename set %s\n", state->filename().latin1());
	note->m_filename = state->filename();
      }
      if (note->isSameFile(m_filepath)) {
	printf("count=%d\n", noteCount);
	noteCount++;
      }
      if (!add(note)) {
	printf("not added\n");
      }
    } else {
      delete note;
    }
  }

  // can't use numNotes() since we can be loading multiple files
  if (m_load_directory_path.isEmpty() && noteCount != m_info.m_num_notes) {
    printf("!!!!! numNotes %d  info says %d !!!!!!\n", noteCount, m_info.m_num_notes);
    assert(noteCount == m_info.m_num_notes);
  }
  printf("have %d notes, directory_path=%s\n", numNotes(), 
	 m_load_directory_path.latin1());
  
  index = 0;
  for(;;) {
    index = xml.find("<paths>",index);
    if (index < 0) {
      break;
    }
    index += 7; // letters in <paths>, points to letter after
    int index2 = xml.find("</paths>",index);
    if (index2 < 0) {
      printf("@@@@ <path> without </paths> at char %d\n", index);
      break;
    }
    if (index2 > index) {
      QStringList newpaths = QStringList::split(QRegExp("[ \t\r\n]+", false, false), 
						xml.mid(index, index2 - index),
						false);
      for(int i=0; i<newpaths.count(); i++) {
	if (!m_paths.contains(newpaths[i])) {
	  m_paths.append(newpaths[i]);
	}
      }
    }
    index = index2;
  }
  printf("Got paths: %s\n", m_paths.join("::").latin1());
  if (m_paths.count() == 0) {
    QFileInfo fi(filepath);
    m_paths.append(fi.dirPath(true));
    printf("no paths, add path = %s\n", fi.dirPath(true).latin1());
  }
  m_locked = false;

  checkdups();

  return(true);
}


QString 
QDocNotes::findPath(const QDocNote &note)
{
  QString path;
  for(int i=0; i<m_paths.count(); i++) {
    path = m_paths[i] + "/" + note.m_filename;
    printf("findPath try: %s\n", path.latin1());
    QFileInfo fi(path);
    if (fi.exists()) {
      printf("found it\n");
      return(path);
    }
  }
  return("");
}

bool 
QDocNote::isSameFile(QString path)
{
  QFileInfo fi(path);
  QString basename = fi.baseName();
  QFileInfo note_fi(m_filename);
  QString note_basename = note_fi.baseName();
  printf("isSameFile %s == note %s\n", basename.latin1(), note_basename.latin1());
  return(basename == note_basename);
}


QDocInfo::QDocInfo()
{
  clear();
}

QDocInfo::~QDocInfo()
{
}

void
QDocInfo::clear()
{
  m_categories.clear();
  m_priority = -1;
  m_num_notes = 0;
  m_pub_date = QDate();
  m_last_open = QDateTime();
  m_last_note = QDateTime();
}  

void
QDocInfo::notesChanged(QDocNotes *notes)
{
  m_last_note = QDateTime::currentDateTime();
}

int
QDocInfo::load(const QString &str, int index)
{
  XmlParser parser;
  QString s;

  index = parser.parse(str, index);
  clear();
  if (!parser.checkHead("info")) {
    //    printf("info header bad\n");
    return(-1);
  }
  int pubdate = 0;

  // mostly have to do all this stupid checking because if you look in
  // qstring.cpp at QString::toLong you will see a bug in the case of strings
  // of length zero  (better hope that stray pointer doesn't point at a plus sign!)

  if (!parser.get("pubdate").isEmpty()) {
    pubdate = parser.get("pubdate").toInt();
  }
  if (pubdate >= 1200 && pubdate < 2200) {
    m_pub_date = QDate(pubdate, 1, 1);
  } else {
    m_pub_date = TimeConversion::fromUTC(pubdate).date();
  }
  if (!parser.get("lastnote").isEmpty()) {
    m_last_note = TimeConversion::fromUTC(parser.get("lastnote").toInt());
  }
  if (!parser.get("lastopen").isEmpty()) {
    m_last_open = TimeConversion::fromUTC(parser.get("lastopen").toInt());
  }
  if (!parser.get("numnotes").isEmpty()) {
    m_num_notes = parser.get("numnotes").toInt();
  }
  if (!parser.get("priority").isEmpty()) {
    m_priority = parser.get("priority").toInt();
  }
  if (!parser.get("categories").isEmpty()) {
    m_categories = QStringList::split(" ",parser.get("categories"));
  }
  //  m_path = str;
  if (qpdf_debug_level > 4) {
    printf("loaded info last note %s last open %s numnotes %d priority %d categories %s\n",
	   m_last_note.toString().latin1(), m_last_open.toString().latin1(),
	   m_num_notes, m_priority, m_categories.join(":").latin1());
  }
  return(index);
}

void
QDocInfo::save(QString &str)
{
  m_last_open = QDateTime::currentDateTime();
  str += "<info ";
  str += " lastopen=\"" + QString::number(TimeConversion::toUTC(m_last_open)) + "\"";
  str += " lastnote=\"" + QString::number(TimeConversion::toUTC(m_last_note)) + "\"";
  str += " pubdate=\"" + QString::number(TimeConversion::toUTC(m_pub_date)) + "\"";
  str += " priority=\"" + QString::number(m_priority) + "\"";
  str += " numnotes=\"" + QString::number(m_num_notes) + "\"";
  str += " categories=\"" + m_categories.join(" ") + "\">\n";
#ifdef QDOCINFO_CACHE
  cache.replace(m_path, new QDocInfo(*this));
#endif
}

bool
QDocInfo::quickRead(QString filepath)
{
  //FIME: this limits the number of categories
  const int header_size = 256;
  char header[header_size];

#ifdef QDOCINFO_CACHE
  if (cache.contains(filepath)) {
    *this = *cache[filepath];
    return true;
  }
#endif

  QFile f( filepath );
  if (qpdf_debug_level > 4) {
    printf("quickRead(%s)\n", filepath.latin1());
  }
  if (filepath.isNull()) {
    return false;
  }

  if ( !f.open(IO_ReadOnly | IO_Raw) ) {
    //    printf("open failed\n");
    return false;
  }

  QString xml;
  int len = f.readBlock(header, header_size);
  if (len > 0) {
    xml.setLatin1(header,len);
  }
  f.close();

  int index;
  index = xml.find("<info ");
  if (index >= 0) {
    load(xml, index);
  }
#ifdef QDOCINFO_CACHE
  cache.insert(filepath, new QDocInfo(*this));
#endif
}

#ifdef QDOCINFO_CACHE
QDocInfoCache QDocInfo::cache;
#endif

