#include "xmlparser.h"
#include <qmap.h>

class QDocNote {
 public:
  friend class QDocNotes;
  QDocNote();
  ~QDocNote();

  void clear();
  void save(QString &str);
  int load(const QString &str, int index);
  int firstWord() { return(m_firstWord); }
  int lastWord() { return(m_lastWord); }
  bool isSameFile(QString path);

  // we remember notes as a range of words, but since the
  // PDF => word algorithm is always being improved, we can't count
  // on word indices to be conserved across versions of the xpdf code.
  // Instead, we remember the start and end coordinates as a ratio of the
  // full size of the page image.  From these and the textmap, we fill in
  // the firstWord, lastWord and regionWords fields.
  int m_page;
  QPoint m_start, m_end;
  int m_firstWord;
  int m_lastWord;
  QString m_filename;

 protected:
  QDateTime m_date;
  QString m_note;
  QRegion m_regionWords;

};

class QDocInfo;

typedef QMap<QString, QDocInfo*> QDocInfoCache;

class QDocInfo {
 public:
  QDocInfo();
  ~QDocInfo();

  void clear();
  bool quickRead(QString path);

  int load(const QString &xml, int index);
  void save(QString &xml);

  void notesChanged(QDocNotes *notes);

  int m_priority;
  int m_num_notes;
  QStringList m_categories;
  QDate m_pub_date;
  QDateTime m_last_note;
  QDateTime m_last_open;
  QString m_path;

 private:
#ifdef DOCINFO_CACHE
  static QDocInfoCache cache;
#endif
};

class QDocNotes {
 public:   
  
  QDocNotes();
  ~QDocNotes();

  void clear();

  void setupPage(QDocText *text);

  // save page position relative to full page size
  // absolutePosition uses current real page size to get real content coord
  QPoint absolutePosition(QPoint pos);
  QPoint relativePosition(QPoint pos);

  QDocNote *find(QDocState &state);
  QDocNote *find(int page, QPoint point);
  QDocNote *find(int page, int word, bool closest);
  QDocNote *find(int page, int firstWord, int lastWord);
  //  QDocNote *find(QString quote);

  QString toText(int page = -1, QDocNote *selected_note = NULL, int *selected_note_position = NULL);
  bool fromText(const QString &text, int noteCharPos, 
		QDocNote **note);

  QRegion highlights(int page);

  QDocNote *add(int page, const QPoint start, const QPoint end, const QString &note, QDocText *text);

  void setPath(const QString &path);
  QString &getPath();
  void setLoadDirectory(const QString &path);
  QString &getLoadDirectory();

  QString findPath(const QDocNote &note);

  static bool loadState(const QString &filepath, QDocState &state);
  bool load(const QString &filepath, QDocState *state);
  bool load(QDocState *state);
  bool save(const QString &filepath, QDocState *state);
  bool save(QDocState *state);

  int numNotes() { return(m_notes.count()); }

  QDocInfo m_info;

  bool isClean() { return(m_clean); }

  bool isLocked() { return(m_locked); }
  void setLocked(bool b) { m_locked = b; }

  void checkdups();

 protected:
  bool add(QDocNote *);
  void updateQuote(QDocNote *);

 private:
  QDocText *m_text;
  QDocState m_setup_state;
  QStringList m_paths;

  QList<QDocNote> m_notes;
  QString m_noteSeperator;
  bool m_clean;

  QString m_filepath;
  QString m_load_directory_path;

  bool m_locked;

  int m_width, m_height;
};

