// Copyright 2003, Philip Kohn

#include "qdocview.h"


QDocText::QDocText()
{
  m_words.setAutoDelete(true);
}


QDocText::~QDocText()
{
  clear();
}

void
QDocText::clear()
{
  m_words.clear();
}

void
QDocText::setText(QString text)
{
  m_text = text;
}

void
QDocText::appendText(QString text)
{
  m_text += text;
}

int
QDocText::numWords()
{
  return(m_words.count());
}

int
QDocText::numChars()
{
  if (m_words.count() <= 0) 
    return(0);
  return(m_words.last()->lastCharIndex);
}

QRect
QDocText::wordBox(int word)
{
  if (word < 0 || word >= numWords()) {
    //    printf("*** wordBox(%d) out of range\n", word);
    return(QRect(0,0,1,1));
  }
  return(m_words.at(word)->box);
}

QRect
QDocText::wordBoxPDF(int word)
{
  if (word < 0 || word >= numWords()) {
    //    printf("*** wordBox(%d) out of range\n", word);
    return(QRect(0,0,1,1));
  }
  return(m_words.at(word)->pdf_box);
}

int
QDocText::findWord(QDocState state, bool closest)
{
  if (isTextMapValid(state)) {
    return(findWord(QPoint(state.x, state.y), closest));
  }
  return(-1);
}
    
int
QDocText::findWord(QPoint point, bool closest, bool lower_right)
{
  // could make this faster with binary search
  QListIterator<QDocTextWord> it(m_words);
  QDocTextWord *word;
  int wordIndex = 0;
  int closestWordDistance = (1<<20);
  int closestWordIndex = -1;
  QPoint d;

  while ((word = it.current()) != 0) {
    ++it;
    if (word->contains(point)) {
      return(wordIndex);
    }
    d = word->box.center() - point;
#if 0
    if (wordIndex < 32) {
      printf("word #%d %s[%d-%d,%d-%d]\n",
	   wordIndex, getText(wordIndex, wordIndex).latin1(),
	   m_words.at(wordIndex)->box.left(),
	   m_words.at(wordIndex)->box.right(),
	   m_words.at(wordIndex)->box.top(),
	     m_words.at(wordIndex)->box.bottom());
    }
#endif
    if (d.manhattanLength() < closestWordDistance) {
      if (!lower_right || (d.x() > 0 && d.y() > 0)) {
	closestWordIndex = wordIndex;
	closestWordDistance = d.manhattanLength();
      }
    }
    wordIndex++;
  }
#if 0
  if (closestWordIndex >= 0) {
    printf("findWord lower_right=%d closest point %d,%d word #%d = %s[%d-%d,%d-%d], dist=%d closest=%d\n",
	   lower_right,
	   point.x(), point.y(), 
	   closestWordIndex, getText(closestWordIndex, closestWordIndex).latin1(),
	   m_words.at(closestWordIndex)->box.left(),
	   m_words.at(closestWordIndex)->box.right(),
	   m_words.at(closestWordIndex)->box.top(),
	   m_words.at(closestWordIndex)->box.bottom(),
	   closestWordDistance, closest);
  }
#endif
  if (!closest) {
    return(-1);
  }
  return(closestWordIndex);
}

int
QDocText::findWordPDF(QPoint point, bool closest)
{
  // could make this faster with binary search
  QListIterator<QDocTextWord> it(m_words);
  QDocTextWord *word;
  int wordIndex = 0;
  int closestWordDistance = (1<<20);
  int closestWordIndex = -1;
  QPoint d;

  while ((word = it.current()) != 0) {
    ++it;
    if (word->containsPDF(point)) {
      return(wordIndex);
    }
    d = word->pdf_box.center() - point;
    if (d.manhattanLength() < closestWordDistance) {
      closestWordIndex = wordIndex;
      closestWordDistance = d.manhattanLength();
    }
    wordIndex++;
  }
#if 0
  if (closestWordIndex >= 0) {
    printf("findWordPDF closest point %d,%d word %d-%d,%d-%d, dist=%d closest=%d\n",
	   point.x(), point.y(), 
	   m_words.at(closestWordIndex)->box.left(),
	   m_words.at(closestWordIndex)->box.right(),
	   m_words.at(closestWordIndex)->box.top(),
	   m_words.at(closestWordIndex)->box.bottom(),
	   closestWordDistance, closest);
  }
#endif
  if (!closest) {
    return(-1);
  }
  return(closestWordIndex);
}

int
QDocText::findWord(QString str)
{
  int charIndex = m_text.find(str);
  if (charIndex < 0) {
    return(-1);
  }
  return(mapCharToWord(charIndex));
}

bool
QDocText::findString(const QString &str, int *firstChar, int *lastChar)
{
  int charIndex;

  if (*firstChar < *lastChar) {
    //    printf("findString (%s) start char %d\n", 
    //	   str.latin1(), *firstChar);
    charIndex = m_text.find(str, *firstChar, false);
    if (charIndex < 0 || charIndex > *lastChar) {
      return(false);
    }
  } else {
    charIndex = m_text.findRev(str, *firstChar, false);
    if (charIndex < 0 || charIndex < *lastChar) {
      return(false);
    }
  }
  *firstChar = charIndex;
  *lastChar = charIndex + str.length() - 1;
  //  printf("found at %d char %d to %d\n", charIndex, *firstChar, *lastChar);
  return(true);
}

QString
QDocText::getText(int firstWord, int lastWord)
{
  if (firstWord <= 0 && lastWord < 0) {
    return(m_text);
  }
  if (lastWord < 0 || lastWord >= numWords()) {
    lastWord = numWords() - 1;
  }
  if (firstWord >= 0 && lastWord >= firstWord) {
    int firstChar, lastChar;
    mapWordToChar(lastWord, &firstChar, &lastChar);
    return(m_text.mid(mapWordToChar(firstWord),
		      lastChar - mapWordToChar(firstWord) + 1));
  } else {
    return QString();
  }
}

int
QDocText::mapCharToWord(int charIndex)
{
  // could make this faster with binary search
  QListIterator<QDocTextWord> it(m_words);
  QDocTextWord *word;
  int wordIndex = 0;

  while ((word = it.current()) != 0) {
    ++it;
    if (word->contains(charIndex)) {
      return(wordIndex);
    }
    wordIndex++;
  }
  return(-1);
}

int
QDocText::mapWordToChar(int wordIndex, int *firstCharIndex, int *lastCharIndex)
{
  if (wordIndex < 0 || wordIndex >= numWords()) {
    if (firstCharIndex) *firstCharIndex = -1;
    if (lastCharIndex) *lastCharIndex = -1;
    return(-1);
  }
  if (firstCharIndex) {
    *firstCharIndex = m_words.at(wordIndex)->firstCharIndex;
  }
  if (lastCharIndex) {
    *lastCharIndex = m_words.at(wordIndex)->lastCharIndex;
  }
  return(m_words.at(wordIndex)->firstCharIndex);
}

		  

		  
QRegion
QDocText::wordRegion(int firstWord, int lastWord)
{
  QRegion region;
  int i;
  if (lastWord < 0 || lastWord >= numWords()) {
    lastWord = numWords() - 1;
  }
  if (firstWord >= 0 && lastWord >= firstWord) {
    for(i = firstWord; i <= lastWord; i++) {
      region |= m_words.at(i)->box;
    }
  }
  return(region);
}

#if 0
QString
QDocText::wordString(int firstWord, int lastWord) 
{
  QString words;
  int i;
  if (lastWord < 0 || lastWord >= numWords()) {
    lastWord = numWords() - 1;
  }
  if (firstWord >= 0 && lastWord >= firstWord) {
    for(i = firstWord; i <= lastWord; i++) {
      words += m_text.mid(m_words.at(i)->firstCharIndex, 
			  m_words.at(i)->lastCharIndex - m_words.at(i)->firstCharIndex + 1);
    }
  }
  return(words);
}
#endif

void
QDocText::addWord(int x1, int y1, int x2, int y2, 
		  int xs1, int xs2, int firstChar, int lastChar)
{
  m_words.append(new QDocTextWord(QRect(x1,y1,x2-x1+1,y2-y1+1), 
				  QRect(xs1,y1,xs2-xs1+1,y2-y1+1),
				  firstChar, lastChar));
#if 0
  printf("add word %d  PDF(%d,%d %d,%d) Screen(%d,%d %d,%d)  chars %d to %d\n",
    	 m_words.count(), x1,y1,x2-x1+1,y2-y1+1, 
	 xs1,y1,xs2-xs1+1,y2-y1+1,firstChar, lastChar);
#endif
}


QDocTextWord::QDocTextWord(QRect _pdf_box, QRect screen_box, int firstChar, int lastChar)
{
  box = screen_box;
  pdf_box = _pdf_box;
  firstCharIndex = firstChar;
  lastCharIndex = lastChar;
}

bool
QDocTextWord::contains(QPoint point)
{
  return(box.contains(point));
}

bool
QDocTextWord::containsPDF(QPoint point)
{
  return(pdf_box.contains(point));
}

bool
QDocTextWord::contains(int charIndex)
{
  return(charIndex >= firstCharIndex && charIndex <= lastCharIndex);
}

